-- This file is a part of Avoision.
-- 
-- Copyright (C) 2011 Jared Krinke.
-- 
-- Avoision is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or (at your option) any later version.
-- 
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

require("Avoision/Common.lua")
require("Avoision/Enemy.lua")
require("Avoision/Seeker.lua")
require("Effects/Ghost.lua")
require("Avoision/Goal.lua")
require("Avoision/Player.lua")

local transitionPeriod = 1000
local boardWidth = 400
local boardHeight = 400
local boardTimeout = 19000
local boardPlayerMinimumDistance = 0.15
local enemyWidth = 1 / 30
local enemyWidthMin = 1 / 90
local enemyWidthMax = 1 / 15
local enemySpeed = 0.2 / 1000
local enemySpeedMin = 0.1 / 1000
local enemySpeedMax = 0.6 / 1000
local enemyGhostPeriod = 500
local enemyGhostScaleMax = 0
local goalWidth = 1 / 30
local goalHeight = 1 / 30
local goalSpeed = 0.1 / 1000
local goalDirectionPeriod = 3000

local pointProgression = { 30, 20, 15, 10, 8, 7, 6, 5, 4, 3, 2, 1, 0 }

Board = { }

function Board.new(layer, lost, ended)
    local board = Entity.new(0, 0, 0, boardWidth, boardHeight)
    
    board.paused = false
    board.finishTimer = 0
    board.difficulty = Common.difficultyLevel.Easy
    board.score = 0
    board.elements:add(Element.Image.new("Images/Background.png", 0, 0, 0, 1, 1, 0))
    board.player = Player.new()
    board.goal = Goal.new(0, 0, goalWidth, goalHeight, 0, 0)
    board.cd = layer:createCollisionDetector()

    -- TODO: Test this more thoroughly
    function board.getSafePosition(self, width, height)
        local ax1 = max(-0.5 + width / 2, self.player.x - (self.player.width / 2 + width / 2 + boardPlayerMinimumDistance))
        local ax2 = min(0.5 - width / 2, self.player.x + (self.player.width / 2 + width / 2 + boardPlayerMinimumDistance))
        local ay1 = max(-0.5 + height / 2, self.player.y - (self.player.height / 2 + height / 2 + boardPlayerMinimumDistance))
        local ay2 = min(0.5 - height / 2, self.player.y + (self.player.height / 2 + height / 2 + boardPlayerMinimumDistance))
    
        repeat
            local x = (random() - 0.5) * (1 - width)
            local y = (random() - 0.5) * (1 - height)
    
            local bx1 = x - width / 2
            local bx2 = x + width / 2
            local by1 = y - height / 2
            local by2 = y + height / 2
    
            local valid = true
    
            if ((ax1 >= bx1 and ax1 <= bx2) or (ax2 >= bx1 and ax2 <= bx2) or (bx1 >= ax1 and bx1 <= ax2))
                and ((ay1 >= by1 and ay1 <= by2) or (ay2 >= by1 and ay2 <= by2) or (by1 >= ay1 and by1 <= ay2))
            then
                valid = false
            else
                return x, y
            end
        until valid
    end
    
    function board.resetGoal(self)
        self:setPoints(pointProgression[1] * self.difficulty)
        self.timer = 0
        
        self.goal.x, self.goal.y = self:getSafePosition(self.goal.width, self.goal.height)

        -- The goal moves on hard
        if self.difficulty >= Common.difficultyLevel.Hard then
            self.goal.axisSpeed = goalSpeed
        else
            self.goal.axisSpeed = 0
        end

        -- Update goal speeds as needed
        self:resetGoalDirection()
    end
    
    function board.addEnemy(self)
        local size = enemyWidth
        local speed = enemySpeed

        -- Enemy speeds vary on normal
        if self.difficulty >= Common.difficultyLevel.Normal then
            speed = enemySpeedMin + (enemySpeedMax - enemySpeedMin) * random()
        end

        -- Enemies are faster on extreme
        if self.difficulty >= Common.difficultyLevel.Extreme then
            speed = speed * 1.1
        end

        -- Enemy sizes vary on hard (and speed varies more)
        if self.difficulty >= Common.difficultyLevel.Hard then
            size = enemyWidthMin + (enemyWidthMax - enemyWidthMin) * random()
        end

        -- Seekers may appear on extreme
        local x = random()
        local seekerProbability = 0

        if self.difficulty >= Common.difficultyLevel.Extreme then
            seekerProbability = 0.15
        end

        local e = nil

        if x >= seekerProbability then
            local speedX = 0
            local speedY = 0

            -- Direction
            if random() > 0.5 then
                speed = -speed
            end

            -- Axis
            if random() > 0.5 then
                speedX = speed
            else
                speedY = speed
            end

            -- On extreme, blocks may move diagonally
            if self.difficulty >= Common.difficultyLevel.Extreme and random() > 0.67 then
                local otherSpeed = enemySpeedMin + (enemySpeedMax - enemySpeedMin) * random()

                if speedX == 0 then
                    speedX = otherSpeed
                elseif speedY == 0 then
                    speedY = otherSpeed
                end
            end

            e = Enemy.new(0, 0, size, size, speedX, speedY)
        else
            e = Seeker.new(0, 0, size, size, speed, self.player)
        end
    
        e.x, e.y = self:getSafePosition(e.width, e.height)

        Ghost.new(e, enemyGhostPeriod, enemyGhostScaleMax, false, function()
            self:addChild(e)
            self.cd:addChild(e)
        end):setParent(self)
    end

    function board.resetGoalDirection(self)
        local speed = self.goal.axisSpeed

        -- Direction
        if random() > 0.5 then
            speed = -speed
        end

        -- Axis
        if random() > 0.5 then
            self.goal.speed.x = speed
            self.goal.speed.y = 0
        else
            self.goal.speed.x = 0
            self.goal.speed.y = speed
        end

        self.goalDirectionTimer = self.timer + goalDirectionPeriod * random()
    end
    
    function board.update(self, ms)
        -- Update children first
        self:updateChildren(ms)

        -- Update timer
        self.timer = self.timer + ms
    
        if not self.paused then
            local done = false

            -- Check for goal intersection
            if self.cd:checkCollision(self.player, self.goal) then
                self:captureGoal()
            else
                -- Update goal direction
                if self.difficulty >= Common.difficultyLevel.Hard and self.timer >= self.goalDirectionTimer then
                    self:resetGoalDirection()
                end
                
                local newPoints = self.difficulty * pointProgression[max(1, min(#pointProgression, 1 + floor(self.timer / boardTimeout * #pointProgression)))]
    
                if newPoints ~= self.points then
                    self:setPoints(newPoints)
                end
            end
    
            -- Check for enemy intersection
            self.cd:forEachCollision(function(p, e)
                done = true
            end, Common.group.player, Common.group.enemy)

            if done then
                self:lose()
            end
        else
            if self.finishTimer > 0 and self.timer >= self.finishTimer then
                ended()
            end
        end
    end
    
    function board.setPoints(self, points)
        self.points = points
    end

    function board.getDifficulty(self)
        return self.difficulty
    end

    function board.setDifficulty(self, difficulty)
        self.difficulty = difficulty
    end
    
    function board.getScore(self)
        return self.score
    end
    
    function board.setScore(self, score)
        self.score = score
    end

    function board.getPaused(self)
        return self.paused
    end
    
    function board.addToScore(self, points)
        self:setScore(self.score + points)
    end

    function board.captureGoal(self)
        self:addToScore(self.points)
        
        -- Create a ghost and move the goal
        self.goal:createGhost():setParent(self)
        self:resetGoal()

        Audio.play("Sounds/Score.wav")

        -- Add another enemy
        self:addEnemy()
    end

    function board.lose(self)
        self.player:clearMovingStates()
        self.player:createGhost():setParent(self)
        self:removeChild(self.player)
        self.paused = true
        self.finishTimer = self.timer + transitionPeriod

        Audio.play("Sounds/Boom.wav")

        lost()
    end
    
    function board.reset(self)
        self.paused = false
        self.finishTimer = 0
        self:setScore(0)
    
        self:clearChildren()
        self.cd:clearChildren()
    
        self:addChild(self.player)
        self:addChild(self.goal)
        self.cd:addChild(self.player)
        self.cd:addChild(self.goal)
    
        self:resetGoal()
    end

    return board
end

