-- This file is a part of Avoision.
-- 
-- Copyright (C) 2012 Jared Krinke.
-- 
-- Avoision is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or (at your option) any later version.
-- 
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

-- TODO: This is a layer and it would make sense to group layers by name or in a folder
require("UI/UI.lua")
require("Avoision/Common.lua")
require("Menus/CreditsMenu.lua")
require("Avoision/GameLayer.lua")
require("Menus/HighScore.lua")
require("Menus/HighScoreEntryMenu.lua")
require("Menus/HighScoreMenu.lua")
require("Menus/InstructionsMenu.lua")
require("Menus/OptionsMenu.lua")
require("Effects/ScriptedEntity.lua")
require("Utility/MiscSettings.lua")

Logo = { }

function Logo.new()
    local logo = Entity.new()

    logo.desiredWidth = 1
    logo.desiredHeight = 1

    -- Animation
    local chaseX = 0.2
    local chaseY = -0.2
    local chaseX2 = 0.8
    local chaseY2 = -0.8
    local chaseSize = 0.25
    local chasePeriod = 3000
    
    logo:addChild(Entity.new(0.5, -0.5, -1, 0.95, 0.95, 0, Color.new(1, 1, 1), ElementList.new(Element.Image.new("Images/Background.png"))))
    logo:addChild(ScriptedEntity.new({ Element.Image.new("Images/Block.png") },
                                     { { 0,               chaseX,  chaseY,  0, chaseSize, chaseSize, 0, Color.new(0, 1, 0) },
                                       { chasePeriod / 4, chaseX2, chaseY,  0, chaseSize, chaseSize, 0, Color.new(0, 1, 0) },
                                       { chasePeriod / 4, chaseX2, chaseY2, 0, chaseSize, chaseSize, 0, Color.new(0, 1, 0) },
                                       { chasePeriod / 4, chaseX,  chaseY2, 0, chaseSize, chaseSize, 0, Color.new(0, 1, 0) },
                                       { chasePeriod / 4, chaseX,  chaseY, 0, chaseSize, chaseSize, 0, Color.new(0, 1, 0) } },
                                     true))
    
    logo:addChild(ScriptedEntity.new({ Element.Image.new("Images/Block.png") },
                                     { { 0,               chaseX2, chaseY,  0, chaseSize, chaseSize, 0, Color.new(1, 0, 0) },
                                       { chasePeriod / 4, chaseX2, chaseY2, 0, chaseSize, chaseSize, 0, Color.new(1, 0, 0) },
                                       { chasePeriod / 4, chaseX,  chaseY2, 0, chaseSize, chaseSize, 0, Color.new(1, 0, 0) },
                                       { chasePeriod / 4, chaseX,  chaseY,  0, chaseSize, chaseSize, 0, Color.new(1, 0, 0) },
                                       { chasePeriod / 4, chaseX2, chaseY,  0, chaseSize, chaseSize, 0, Color.new(1, 0, 0) } },
                                     true))
    
    logo:addChild(ScriptedEntity.new({ Element.Image.new("Images/Block.png") },
                                     { { 0,               chaseX,  chaseY2, 0, chaseSize, chaseSize, 0, Color.new(1, 1, 1) },
                                       { chasePeriod / 4, chaseX,  chaseY,  0, chaseSize, chaseSize, 0, Color.new(1, 1, 1) },
                                       { chasePeriod / 4, chaseX2, chaseY,  0, chaseSize, chaseSize, 0, Color.new(1, 1, 1) },
                                       { chasePeriod / 4, chaseX2, chaseY2, 0, chaseSize, chaseSize, 0, Color.new(1, 1, 1) },
                                       { chasePeriod / 4, chaseX,  chaseY2, 0, chaseSize, chaseSize, 0, Color.new(1, 1, 1) } },
                                     true))

    -- Component implementation
    function logo.setLayer(self, layer)
        layer:addChild(self)
    end

    function logo.getPosition(self)
        return self.x, self.y
    end

    function logo.setPosition(self, x, y)
        self.x = x
        self.y = y
    end

    function logo.getActive(self)
        return false
    end

    function logo.getMinimumSize(self)
        return self.desiredWidth, self.desiredHeight
    end

    logo.getDesiredSize = logo.getMinimumSize
    logo.getSize = logo.getMinimumSize

    function logo.setSize(self, width, height)
        -- Maintain aspect ratio
        self.desiredWidth = max(width, height)
        self.desiredHeight = self.desiredWidth

        self.width = self.desiredWidth
        self.height = self.desiredWidth
    end

    return logo
end

MainMenu = { }

function MainMenu.new()
    local mainForm = Form.newGrid(1, -200, 200, 400)

    mainForm.gameLayer = GameLayer.new(function(difficulty, score)
        if score > HighScore.getLowestScore(difficulty) then
            mainForm:showHighScoreEntry(difficulty, score, function() mainForm:showHighScores() end)
        else
            mainForm:showHighScores()
        end
    end)

    mainForm.instructionsMenu = InstructionsMenu.new()
    mainForm.creditsMenu = CreditsMenu.new()
    mainForm.highScoreEntryMenu = HighScoreEntryMenu.new()
    mainForm.highScoreMenu = HighScoreMenu.new()
    mainForm.optionsMenu = OptionsMenu.new()

    function mainForm.startNewGame(self, showInstructions)
        self.gameLayer:setDifficulty(self.difficulty)
        self.gameLayer:reset()
        Layers.push(self.gameLayer)
        self.gameLayer:started()

        if showInstructions then
            self:showInstructions(false)
        end
    end

    function mainForm.showInstructions(self, propagateAudio)
        self.instructionsMenu.propagateAudio = propagateAudio
        self.instructionsMenu:show()
    end

    function mainForm.showHighScoreEntry(self, difficulty, score, entered)
        self.highScoreEntryMenu:show(difficulty, score, entered)
    end

    function mainForm.showHighScores(self)
        self.highScoreMenu:show()
    end

    function mainForm.showVideoOptions(self)
        self.videoMenu:show()
    end
    
    function mainForm.exit(self)
        exit()
    end

    -- Form
    local difficultyChoice = Choice.new("Difficulty", Common.difficultyLevelNames, function(self, choice) mainForm.difficulty = Common.difficultyLevel[choice] end)

    mainForm:add(Form.newNestedCenterFlow(3,
        ImageLabel.new("Images/Title.png", Common.textWidth * 1.9 * 8, Common.textHeight * 1.75, "center"),
        Label.new("  "),
        Logo.new()))

    mainForm:add(Form.newNestedLeftFlow(1,
        Separator.new(),
        Button.new("New Game", function()
            -- Save difficulty selection (and only show instructions by default the first time)
            local showInstructions = MiscSettings.get("boolean", "showInstructions", true)

            MiscSettings.set("boolean", "showInstructions", false)
            MiscSettings.set("string", "difficulty", Common.difficultyLevelNames[mainForm.difficulty])
            MiscSettings.save()

            -- Start game
            mainForm:startNewGame(showInstructions)
        end, true),

        difficultyChoice,
        Button.new("High Scores", function() mainForm:showHighScores() end),
        Separator.new(),
        Button.new("How to Play", function() mainForm:showInstructions(true) end),
        Button.new("Options", function() mainForm.optionsMenu:show()  end),
        Button.new("Credits", function() mainForm.creditsMenu:show() end, true),
        Separator.new(),
        Button.new("Quit", function() mainForm:exit() end)))

    -- Default to initial difficulty
    local initialDifficulty = MiscSettings.get("string", "difficulty")

    if isString(initialDifficulty) then
        difficultyChoice:select(initialDifficulty)
    end

    -- Center form vertically
    local x, y = mainForm:getPosition()
    local width, height = mainForm:getMinimumSize()

    mainForm:setPosition(x, height / 2)

    -- Create layer
    local mainMenu = FormLayer.new(mainForm, 33)

    -- Version
    local version = Entity.new(-315, -235, 0, Common.textWidth / 2, Common.textHeight / 2)

    version.elements:add(Element.Text.new("v" .. Common.version))
    version.color = Color.new(0.5, 0.5, 0.5)
    mainMenu:addChild(version)
    
    -- Error handling
    function mainMenu.errorOccurred(self, message)
        print(message)
    end

    return mainMenu
end

